#!/usr/bin/python


# BrraWrapper module
import os
import urllib2
import urllib
import cookielib
import tempfile
import time
import HTMLParser
import base64
import ConfigParser
import socket
import uuid
import syslog

# log levels
LOG_NONE = 'none'
LOG_ERROR = 'error'
LOG_DEBUG = 'debug'

# custom exception class
class BrraWrapperError(Exception): pass

# public.brra.bg changed to portal.registryagency.bg
# configuration defaults
defaults = {
	'HttpTimeout'		: '30',
	'DefaultUrl'		: 'https://portal.registryagency.bg/CheckUps/Default.ra',
	'AuthenticationUrl'	: 'https://portal.registryagency.bg/Common/AuthenticationPage.ra',
	'CookieName'		: 'ASP.NET_SessionId',
	'CertificateFile'	: '24422387.p12',
	'CertificatePin'	: '24422387',
	'CheckGuid'		: 'f60bb2709d944c88b6cd5edf29717ef6',
	#'AuthCheckUrl'		: 'https://portal.registryagency.bg/Documents/RetrieveDocument.ra?DocGuid=f60bb2709d944c88b6cd5edf29717ef6',
	'AuthCheckUrl'		: 'https://portal.registryagency.bg/CR/DocumentAccess/8cbfcd8bba814b9991d4a31010a1ef00',
	'XmlSecBinary'		: '/usr/bin/xmlsec1',
	'LogLevel'		: LOG_ERROR,
}

# configuration parser
config = ConfigParser.ConfigParser(defaults)
config.read(['/etc/BrraWrapper.cfg', '/etc/BrraWrapper/BrraWrapper.cfg', 'BrraWrapper.cfg'])


# log message to system log
def LogMessage(prio, message):
	LogLevel = config.get('BrraWrapper', 'LogLevel')
	if LogLevel == LOG_NONE:
		return
	if LogLevel == LOG_DEBUG or prio == syslog.LOG_ERR:
		syslog.openlog('BrraWrapper', 0, syslog.LOG_USER)
		syslog.syslog(prio, message)
		syslog.closelog()


# XML template file
XmlTemplate = '''\
<?xml version="1.0" encoding="UTF-8"?>
<AuthorizationSignature>
   <Signature xmlns="http://www.w3.org/2000/09/xmldsig#" Id="_%(GUID1)s">
      <SignedInfo>
         <CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315#WithComments" />
         <SignatureMethod Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1" />
         <Reference URI="">
            <Transforms>
               <Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature" />
            </Transforms>
            <DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1" />
            <DigestValue></DigestValue>
         </Reference>
         <Reference URI="#_%(GUID2)s">
            <DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1" />
            <DigestValue></DigestValue>
         </Reference>
      </SignedInfo>
      <SignatureValue></SignatureValue>
      <KeyInfo>
         <X509Data>
            <X509Certificate>
            </X509Certificate>
         </X509Data>
         <KeyValue>
            <RSAKeyValue>
               <Modulus></Modulus>
               <Exponent></Exponent>
            </RSAKeyValue>
         </KeyValue>
      </KeyInfo>
      <Object>
         <xades:QualifyingProperties xmlns:xades="http://uri.etsi.org/01903/v1.3.2#" Target="#_%(GUID1)s">
            <xades:SignedProperties Id="_%(GUID2)s">
               <xades:SignedSignatureProperties>
                  <xades:SigningTime>%(DateTime)s</xades:SigningTime>
               </xades:SignedSignatureProperties>
            </xades:SignedProperties>
         </xades:QualifyingProperties>
      </Object>
   </Signature>
</AuthorizationSignature>
'''

# sign XML template file and return result string
def SignXML(XmlTemplate):
	# command to use for xml signing
	cmd = '%(XmlSecBinary)s --sign --pkcs12 %(Certificate)s --pwd %(Password)s %(TemplateFile)s'
	# write template to the file
	LogMessage(syslog.LOG_INFO, 'SignXML(): Creating temporary file')
	fd, TemplateFile = tempfile.mkstemp()
	fh = os.fdopen(fd, 'w+')
	Data = XmlTemplate % {
		'GUID1': uuid.uuid4(),
		'GUID2': uuid.uuid4(),
		'DateTime': time.strftime('%Y-%m-%dT%H:%M:%S%z'),
	}
	LogMessage(syslog.LOG_INFO, 'SignXML(): Writing XML template to temporary file')
	fh.write(Data)
	fh.close()
	# sign with pipe and read output
	command = cmd % {
		'XmlSecBinary': config.get('BrraWrapper', 'XmlSecBinary'),
		'Certificate': config.get('BrraWrapper', 'CertificateFile'),
		'Password': config.get('BrraWrapper', 'CertificatePin'),
		'TemplateFile': TemplateFile,
	}
	LogMessage(syslog.LOG_INFO, 'SignXML(): Signing XML template in temporary file')
	pipe = os.popen(command)
	Signed = pipe.read()
	xvr = pipe.close()
	# clean up temp directory
	LogMessage(syslog.LOG_INFO, 'SignXML(): Cleaning temporary file')
	os.unlink(TemplateFile)
	if xvr:
		raise BrraWrapperError('XmlSec1 error, check certificate or PIN')
	return Signed


# enforce http request timeouts 
def SafeTimeout(callback):
	def decorator(*args):
		# set desired timeout
		timeout = socket.getdefaulttimeout()
		socket.setdefaulttimeout(float(config.get('BrraWrapper', 'HttpTimeout')))
		try:
			# run 
			res = callback(*args)
		except:
			# restore default timeout in case of error
			# and let the error be handled elsewhere
			socket.setdefaulttimeout(timeout)
			raise
		# restore default timeout
		socket.setdefaulttimeout(timeout)
		return res
	return decorator


# log and possibly handle errors
def ErrorHandler(callback):
	def decorator(*args):
		try:
			LogMessage(syslog.LOG_INFO, 'ErrorHandler(): About to enter %s' % callback.__name__)
			res = callback(*args)
			LogMessage(syslog.LOG_INFO, 'ErrorHandler(): %s exited with %s' % (callback.__name__, res))
			return res
		except socket.timeout:
			LogMessage(syslog.LOG_ERR, 'ErrorHandler(): ERROR: Timeout occured')
			return 2
		except BrraWrapperError, e:
			LogMessage(syslog.LOG_ERR, 'ErrorHandler(): ERROR: %s' % str(e))
			return 3
		except Exception, e:
			#if len(e)==0: e=["Unhandled exception"]
			LogMessage(syslog.LOG_ERR, 'ErrorHandler(): ERROR: Unhandled exception: %s' % str(e[0]))
			raise
	return decorator


# open and close syslog
def LogHandler(callback):
	def decorator(*args):
		syslog.openlog('BrraWrapper', 0, syslog.LOG_USER)
		mask = syslog.setlogmask(syslog.LOG_UPTO(syslog.LOG_INFO))
		res = callback(*args)
		syslog.closelog()
		syslog.setlogmask(mask)
		return res
	return decorator


# bundle all decorators together
def bundle(callback):
	@LogHandler
	@ErrorHandler
	@SafeTimeout
	def decorator(*args):
		return callback(*args)
	return decorator


# get new session id
@bundle
def GetNewSession():
	cj = cookielib.CookieJar()
	opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(cj))
	LogMessage(syslog.LOG_INFO, 'GetNewSession(): Preparing request object')
	req = urllib2.Request(config.get('BrraWrapper', 'DefaultUrl'))
	req.get_method = lambda: 'HEAD'
	HTTPTimeout = 30
	LogMessage(syslog.LOG_INFO, 'GetNewSession(): Open URL')
	res = opener.open(req)
	for cookie in cj:
		if cookie.name == config.get('BrraWrapper', 'CookieName'):
			return cookie.value
	raise BrraWrapperError('Unable to get new session!')


# authenticate session
@bundle
def AuthenticateSession(SessionId):
	# parse HTML data
	class MyParser(HTMLParser.HTMLParser):
		data = {}
		element = None

		# get important data
		def handle_starttag(self, tag, attrs):
			FormFields = (
				'PageRefreshTicketIdInHiddenField',
				'__VIEWSTATE',
				'__EVENTVALIDATION',
				'UcSign$AppletRetStatus',
				'UcSign$communication_message',
				'UcSign$data',
				'UcSign$post_sign_data',
			)

			data = {}
			# save data in temporary variable
			for attrName, attrValue in attrs:
				data[attrName] = attrValue
			# if required attribute is in the data - save it for later reference
			if 'name' in data and data['name'] in FormFields:
				self.data[data['name']] = data.get('value', '')
	
	# prepare cookie data
	cookie = '%(CookieName)s=%(CookieValue)s' % {'CookieName': config.get('BrraWrapper', 'CookieName'), 'CookieValue': SessionId}
	# prepare request object
	LogMessage(syslog.LOG_INFO, 'AuthenticateSession(): Preparing request object')
	req = urllib2.Request(config.get('BrraWrapper', 'AuthenticationUrl'))
	req.add_header('Cookie', cookie)

	# run first request
	LogMessage(syslog.LOG_INFO, 'AuthenticateSession(): Open URL')
	res = urllib2.urlopen(req)

	# parse HTML document
	parser = MyParser()
	LogMessage(syslog.LOG_INFO, 'AuthenticateSession(): Parsing HTML')
	parser.feed(res.read())
	parser.close()

	# put signature xml in the form
	LogMessage(syslog.LOG_INFO, 'AuthenticateSession(): Preparing request object (authenticate)')
	parser.data['UcSign$post_sign_data'] = base64.b64encode(SignXML(XmlTemplate))

	# run second authentication request
	LogMessage(syslog.LOG_INFO, 'AuthenticateSession(): Open URL (authenticate)')
	res = urllib2.urlopen(req, data=urllib.urlencode(parser.data)).geturl() == config.get('BrraWrapper', 'AuthenticationUrl')

	LogMessage(syslog.LOG_INFO, 'AuthenticateSession(): finished with %s value' % str(res))
	return res


# returns true if session is valid, false otherwise
@bundle
def IsValidSession(SessionId):
	cookie = '%(CookieName)s=%(CookieValue)s' % {'CookieName': config.get('BrraWrapper', 'CookieName'), 'CookieValue': SessionId}
	cj = cookielib.CookieJar()
	opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(cj))
	LogMessage(syslog.LOG_INFO, 'IsValidSession(): Preparing request object')
	req = urllib2.Request(config.get('BrraWrapper', 'DefaultUrl'))
	req.get_method = lambda: 'HEAD'
	req.add_header('Cookie', cookie)
	LogMessage(syslog.LOG_INFO, 'IsValidSession(): Open URL')
	opener.open(req)
	res = True
	for cookie in cj:
		if cookie.name == config.get('BrraWrapper', 'CookieName'):
			res = cookie.value == SessionId
	LogMessage(syslog.LOG_INFO, 'IsValidSession(): finished with %s value' % str(res))
	return res


# check session authentication status
@bundle
def IsAuthenticatedSession(SessionId):
	cookie = '%(CookieName)s=%(CookieValue)s' % {'CookieName': config.get('BrraWrapper', 'CookieName'), 'CookieValue': SessionId}
	LogMessage(syslog.LOG_INFO, 'IsAuthenticatedSession(): Preparing request object')
	req = urllib2.Request(config.get('BrraWrapper', 'AuthCheckUrl'))
	req.get_method = lambda: 'HEAD'
	req.add_header('Cookie', cookie)
	LogMessage(syslog.LOG_INFO, 'IsAuthenticatedSession(): Open URL')
	res = urllib2.urlopen(req).geturl() == config.get('BrraWrapper', 'AuthCheckUrl')
	LogMessage(syslog.LOG_INFO, 'IsAuthenticatedSession(): finished with %s value' % str(res))
	return res


# main program (in case module is used as a standalone program)
if __name__ == '__main__':
	import optparse
	parser = optparse.OptionParser('Usage: %prog [options] [arg]')
	parser.add_option('-n', '--new', dest='newsid', action='store_true', help='Get new SessionID', default=False)
	parser.add_option('-a', '--authenticate', dest='authenticate', help='Authenticate session', metavar='SessionId')
	parser.add_option('-v', '--is-valid', dest='isvalid', help='Check if SessionID is valid', metavar='SessionId')
	parser.add_option('-t', '--is-authenticated', dest='isauth', help='Check if SessionID is authenticated', metavar='SessionId')
	(options, args) = parser.parse_args()
	
	if options.newsid:
		print GetNewSession()
	elif options.authenticate:
		print AuthenticateSession(options.authenticate)
	elif options.isvalid:
		print IsValidSession(options.isvalid)
	elif options.isauth:
		print IsAuthenticatedSession(options.isauth)
	else:
		parser.error('use -h for help')
